# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
Django production settings for swh-web.
"""

from .common import * # noqa
from .common import (
    MIDDLEWARE, CACHES, ALLOWED_HOSTS, WEBPACK_LOADER
)
from .common import swh_web_config
from .common import REST_FRAMEWORK

# activate per-site caching
if 'GZip' in MIDDLEWARE[0]:
    MIDDLEWARE.insert(1, 'django.middleware.cache.UpdateCacheMiddleware')
else:
    MIDDLEWARE.insert(0, 'django.middleware.cache.UpdateCacheMiddleware')

MIDDLEWARE += ['swh.web.common.middlewares.HtmlMinifyMiddleware',
               'django.middleware.cache.FetchFromCacheMiddleware']

CACHES.update({
    'default': {
        'BACKEND': 'django.core.cache.backends.memcached.MemcachedCache',
        'LOCATION': swh_web_config['throttling']['cache_uri'],
    }
})

# Setup support for proxy headers
USE_X_FORWARDED_HOST = True
SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')

# We're going through seven (or, in that case, 2) proxies thanks to Varnish
REST_FRAMEWORK['NUM_PROXIES'] = 2

ALLOWED_HOSTS += [
    'archive.softwareheritage.org',
    'base.softwareheritage.org',
    'archive.internal.softwareheritage.org',
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': swh_web_config['production_db'],
    }
}

WEBPACK_LOADER['DEFAULT']['CACHE'] = True



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('swh.web.common', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='saveoriginrequest',
            name='visit_date',
            field=models.DateTimeField(null=True),
        ),
    ]



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from urllib.parse import unquote


from django.contrib.auth import get_user_model
from unittest.mock import patch

from swh.web.common.models import (
    SaveAuthorizedOrigin, SaveUnauthorizedOrigin
)
from swh.web.common.origin_save import can_save_origin
from swh.web.common.models import (
    SAVE_REQUEST_PENDING, SAVE_REQUEST_ACCEPTED,
    SAVE_REQUEST_REJECTED, SAVE_TASK_NOT_YET_SCHEDULED
)
from swh.web.common.utils import reverse
from swh.web.tests.testcase import WebTestCase

_user_name = 'swh-web-admin'
_user_mail = 'admin@swh-web.org'
_user_password = '..34~pounds~BEAUTY~march~63..'

_authorized_origin_url = 'https://scm.ourproject.org/anonscm/'
_unauthorized_origin_url = 'https://www.softwareheritage.org/'


class OriginSaveAdminTestCase(WebTestCase):

    @classmethod
    def setUpTestData(cls):  # noqa: N802
        User = get_user_model()  # noqa: N806
        user = User.objects.create_user(_user_name, _user_mail, _user_password)
        user.is_staff = True
        user.save()
        SaveAuthorizedOrigin.objects.create(url=_authorized_origin_url)
        SaveUnauthorizedOrigin.objects.create(url=_unauthorized_origin_url)

    def check_not_login(self, url):
        login_url = reverse('login', query_params={'next': url})
        response = self.client.post(url)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(unquote(response.url), login_url)

    def test_add_authorized_origin_url(self):
        authorized_url = 'https://scm.adullact.net/anonscm/'
        self.assertEqual(can_save_origin(authorized_url),
                         SAVE_REQUEST_PENDING)

        url = reverse('admin-origin-save-add-authorized-url',
                      url_args={'origin_url': authorized_url})

        self.check_not_login(url)

        self.assertEqual(can_save_origin(authorized_url),
                         SAVE_REQUEST_PENDING)

        self.client.login(username=_user_name, password=_user_password)
        response = self.client.post(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(can_save_origin(authorized_url),
                         SAVE_REQUEST_ACCEPTED)

    def test_remove_authorized_origin_url(self):
        self.assertEqual(can_save_origin(_authorized_origin_url),
                         SAVE_REQUEST_ACCEPTED)

        url = reverse('admin-origin-save-remove-authorized-url',
                      url_args={'origin_url': _authorized_origin_url})

        self.check_not_login(url)

        self.assertEqual(can_save_origin(_authorized_origin_url),
                         SAVE_REQUEST_ACCEPTED)

        self.client.login(username=_user_name, password=_user_password)
        response = self.client.post(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(can_save_origin(_authorized_origin_url),
                         SAVE_REQUEST_PENDING)

    def test_add_unauthorized_origin_url(self):
        unauthorized_url = 'https://www.yahoo./'
        self.assertEqual(can_save_origin(unauthorized_url),
                         SAVE_REQUEST_PENDING)

        url = reverse('admin-origin-save-add-unauthorized-url',
                      url_args={'origin_url': unauthorized_url})

        self.check_not_login(url)

        self.assertEqual(can_save_origin(unauthorized_url),
                         SAVE_REQUEST_PENDING)

        self.client.login(username=_user_name, password=_user_password)
        response = self.client.post(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(can_save_origin(unauthorized_url),
                         SAVE_REQUEST_REJECTED)

    def test_remove_unauthorized_origin_url(self):
        self.assertEqual(can_save_origin(_unauthorized_origin_url),
                         SAVE_REQUEST_REJECTED)

        url = reverse('admin-origin-save-remove-unauthorized-url',
                      url_args={'origin_url': _unauthorized_origin_url})

        self.check_not_login(url)

        self.assertEqual(can_save_origin(_unauthorized_origin_url),
                         SAVE_REQUEST_REJECTED)

        self.client.login(username=_user_name, password=_user_password)
        response = self.client.post(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(can_save_origin(_unauthorized_origin_url),
                         SAVE_REQUEST_PENDING)

    @patch('swh.web.common.origin_save.scheduler')
    def test_accept_pending_save_request(self, mock_scheduler):
        origin_type = 'git'
        origin_url = 'https://v2.pikacode.com/bthate/botlib.git'
        save_request_url = reverse('api-save-origin',
                                   url_args={'origin_type': origin_type,
                                             'origin_url': origin_url})
        response = self.client.post(save_request_url, data={},
                                    content_type='application/x-www-form-urlencoded') # noqa
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['save_request_status'],
                         SAVE_REQUEST_PENDING)

        accept_request_url = reverse('admin-origin-save-request-accept',
                                     url_args={'origin_type': origin_type,
                                               'origin_url': origin_url})

        self.check_not_login(accept_request_url)

        tasks_data = [
            {
                'priority': 'high',
                'policy': 'oneshot',
                'type': 'origin-update-git',
                'arguments': {
                    'kwargs': {
                        'repo_url': origin_url
                    },
                    'args': []
                },
                'status': 'next_run_not_scheduled',
                'id': 1,
             }
        ]

        mock_scheduler.create_tasks.return_value = tasks_data
        mock_scheduler.get_tasks.return_value = tasks_data

        self.client.login(username=_user_name, password=_user_password)
        response = self.client.post(accept_request_url)
        self.assertEqual(response.status_code, 200)

        response = self.client.get(save_request_url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data[0]['save_request_status'],
                         SAVE_REQUEST_ACCEPTED)
        self.assertEqual(response.data[0]['save_task_status'],
                         SAVE_TASK_NOT_YET_SCHEDULED)

    @patch('swh.web.common.origin_save.scheduler')
    def test_reject_pending_save_request(self, mock_scheduler):
        origin_type = 'git'
        origin_url = 'https://wikipedia.com'
        save_request_url = reverse('api-save-origin',
                                   url_args={'origin_type': origin_type,
                                             'origin_url': origin_url})
        response = self.client.post(save_request_url, data={},
                                    content_type='application/x-www-form-urlencoded') # noqa
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['save_request_status'],
                         SAVE_REQUEST_PENDING)

        reject_request_url = reverse('admin-origin-save-request-reject',
                                     url_args={'origin_type': origin_type,
                                               'origin_url': origin_url})

        self.check_not_login(reject_request_url)

        self.client.login(username=_user_name, password=_user_password)
        response = self.client.post(reject_request_url)
        self.assertEqual(response.status_code, 200)

        tasks_data = [
            {
                'priority': 'high',
                'policy': 'oneshot',
                'type': 'origin-update-git',
                'arguments': {
                    'kwargs': {
                        'repo_url': origin_url
                    },
                    'args': []
                },
                'status': 'next_run_not_scheduled',
                'id': 1,
             }
        ]

        mock_scheduler.create_tasks.return_value = tasks_data
        mock_scheduler.get_tasks.return_value = tasks_data

        response = self.client.get(save_request_url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data[0]['save_request_status'],
                         SAVE_REQUEST_REJECTED)



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random

from hypothesis import given

from swh.web.common.utils import reverse, get_swh_persistent_id
from swh.web.common.utils import gen_path_info
from swh.web.tests.strategies import (
    directory, directory_with_subdirs, invalid_sha1,
    unknown_directory
)
from swh.web.tests.testcase import WebTestCase


class SwhBrowseDirectoryTest(WebTestCase):

    def directory_view(self, root_directory_sha1, directory_entries,
                       path=None):
        dirs = [e for e in directory_entries if e['type'] in ('dir', 'rev')]
        files = [e for e in directory_entries if e['type'] == 'file']

        url_args = {'sha1_git': root_directory_sha1}
        if path:
            url_args['path'] = path

        url = reverse('browse-directory',
                      url_args=url_args)

        root_dir_url = reverse('browse-directory',
                               url_args={'sha1_git': root_directory_sha1})

        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed('browse/directory.html')
        self.assertContains(resp, '<a href="' + root_dir_url + '">' +
                            root_directory_sha1[:7] + '</a>')
        self.assertContains(resp, '<td class="swh-directory">',
                            count=len(dirs))
        self.assertContains(resp, '<td class="swh-content">',
                            count=len(files))

        for d in dirs:
            if d['type'] == 'rev':
                dir_url = reverse('browse-revision',
                                  url_args={'sha1_git': d['target']})
            else:
                dir_path = d['name']
                if path:
                    dir_path = "%s/%s" % (path, d['name'])
                dir_url = reverse('browse-directory',
                                  url_args={'sha1_git': root_directory_sha1,
                                            'path': dir_path})
            self.assertContains(resp, dir_url)

        for f in files:
            file_path = "%s/%s" % (root_directory_sha1, f['name'])
            if path:
                file_path = "%s/%s/%s" % (root_directory_sha1, path, f['name'])
            query_string = 'sha1_git:' + f['target']
            file_url = reverse('browse-content',
                               url_args={'query_string': query_string},
                               query_params={'path': file_path})
            self.assertContains(resp, file_url)

        path_info = gen_path_info(path)

        self.assertContains(resp, '<li class="swh-path">',
                            count=len(path_info)+1)
        self.assertContains(resp, '<a href="%s">%s</a>' %
                                  (root_dir_url, root_directory_sha1[:7]))

        for p in path_info:
            dir_url = reverse('browse-directory',
                              url_args={'sha1_git': root_directory_sha1,
                                        'path': p['path']})
            self.assertContains(resp, '<a href="%s">%s</a>' %
                                      (dir_url, p['name']))

        self.assertContains(resp, 'vault-cook-directory')

        swh_dir_id = get_swh_persistent_id('directory', directory_entries[0]['dir_id']) # noqa
        swh_dir_id_url = reverse('browse-swh-id',
                                 url_args={'swh_id': swh_dir_id})
        self.assertContains(resp, swh_dir_id)
        self.assertContains(resp, swh_dir_id_url)

    @given(directory())
    def test_root_directory_view(self, directory):
        self.directory_view(directory, self.directory_ls(directory))

    @given(directory_with_subdirs())
    def test_sub_directory_view(self, directory):
        dir_content = self.directory_ls(directory)
        subdir = random.choice([e for e in dir_content if e['type'] == 'dir'])
        subdir_content = self.directory_ls(subdir['target'])
        self.directory_view(directory, subdir_content, subdir['name'])

    @given(invalid_sha1(), unknown_directory())
    def test_directory_request_errors(self, invalid_sha1, unknown_directory):

        dir_url = reverse('browse-directory',
                          url_args={'sha1_git': invalid_sha1})

        resp = self.client.get(dir_url)
        self.assertEqual(resp.status_code, 400)
        self.assertTemplateUsed('browse/error.html')

        dir_url = reverse('browse-directory',
                          url_args={'sha1_git': unknown_directory})

        resp = self.client.get(dir_url)
        self.assertEqual(resp.status_code, 404)
        self.assertTemplateUsed('browse/error.html')



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from rest_framework.decorators import api_view

from swh.web.common.urlsindex import UrlsIndex
from swh.web.common.throttling import throttle_scope


class APIUrls(UrlsIndex):
    """
    Class to manage API documentation URLs.

    - Indexes all routes documented using apidoc's decorators.
    - Tracks endpoint/request processing method relationships for use in
      generating related urls in API documentation

    """
    _apidoc_routes = {}
    _method_endpoints = {}
    scope = 'api'

    @classmethod
    def get_app_endpoints(cls):
        return cls._apidoc_routes

    @classmethod
    def add_route(cls, route, docstring, **kwargs):
        """
        Add a route to the self-documenting API reference
        """
        route_view_name = 'api-%s' % route[1:-1].replace('/', '-')
        if route not in cls._apidoc_routes:
            d = {'docstring': docstring,
                 'route_view_name': route_view_name}
            for k, v in kwargs.items():
                d[k] = v
            cls._apidoc_routes[route] = d


class api_route(object):  # noqa: N801
    """
    Decorator to ease the registration of an API endpoint
    using the Django REST Framework.

    Args:
        url_pattern: the url pattern used by DRF to identify the API route
        view_name: the name of the API view associated to the route used to
           reverse the url
        methods: array of HTTP methods supported by the API route

    """

    def __init__(self, url_pattern=None, view_name=None,
                 methods=['GET', 'HEAD', 'OPTIONS'],
                 throttle_scope='swh_api',
                 api_version='1'):
        super().__init__()
        self.url_pattern = '^' + api_version + url_pattern + '$'
        self.view_name = view_name
        self.methods = methods
        self.throttle_scope = throttle_scope

    def __call__(self, f):
        # create a DRF view from the wrapped function
        @api_view(self.methods)
        @throttle_scope(self.throttle_scope)
        def api_view_f(*args, **kwargs):
            return f(*args, **kwargs)
        # small hacks for correctly generating API endpoints index doc
        api_view_f.__name__ = f.__name__
        api_view_f.http_method_names = self.methods

        # register the route and its view in the endpoints index
        APIUrls.add_url_pattern(self.url_pattern, api_view_f,
                                self.view_name)
        return f



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timedelta
from django.utils import timezone

from rest_framework.test import APITestCase
from unittest.mock import patch

from swh.web.common.utils import reverse
from swh.web.common.models import (
    SaveUnauthorizedOrigin, SaveOriginRequest,
    SAVE_REQUEST_ACCEPTED, SAVE_REQUEST_REJECTED,
    SAVE_REQUEST_PENDING
)
from swh.web.common.models import (
    SAVE_TASK_NOT_CREATED, SAVE_TASK_NOT_YET_SCHEDULED,
    SAVE_TASK_SCHEDULED, SAVE_TASK_FAILED, SAVE_TASK_SUCCEED
)
from swh.web.tests.testcase import WebTestCase


class SaveApiTestCase(WebTestCase, APITestCase):

    @classmethod
    def setUpTestData(cls):  # noqa: N802
        SaveUnauthorizedOrigin.objects.create(
            url='https://github.com/user/illegal_repo')
        SaveUnauthorizedOrigin.objects.create(
            url='https://gitlab.com/user_to_exclude')

    def test_invalid_origin_type(self):
        url = reverse('api-save-origin',
                      url_args={'origin_type': 'foo',
                                'origin_url': 'https://github.com/torvalds/linux'}) # noqa

        response = self.client.post(url)
        self.assertEqual(response.status_code, 400)

    def test_invalid_origin_url(self):
        url = reverse('api-save-origin',
                      url_args={'origin_type': 'git',
                                'origin_url': 'bar'})

        response = self.client.post(url)
        self.assertEqual(response.status_code, 400)

    def check_created_save_request_status(self, mock_scheduler, origin_url,
                                          scheduler_task_status,
                                          expected_request_status,
                                          expected_task_status=None,
                                          visit_date=None):

        if not scheduler_task_status:
            mock_scheduler.get_tasks.return_value = []
        else:
            mock_scheduler.get_tasks.return_value = \
                [{
                    'priority': 'high',
                    'policy': 'oneshot',
                    'type': 'origin-update-git',
                    'arguments': {
                        'kwargs': {
                            'repo_url': origin_url
                        },
                        'args': []
                    },
                    'status': scheduler_task_status,
                    'id': 1,
                }]

        mock_scheduler.create_tasks.return_value = \
            [{
                'priority': 'high',
                'policy': 'oneshot',
                'type': 'origin-update-git',
                'arguments': {
                    'kwargs': {
                        'repo_url': origin_url
                    },
                    'args': []
                },
                'status': 'next_run_not_scheduled',
                'id': 1,
             }]

        url = reverse('api-save-origin',
                      url_args={'origin_type': 'git',
                                'origin_url': origin_url})

        with patch('swh.web.common.origin_save._get_visit_info_for_save_request') as mock_visit_date: # noqa
            mock_visit_date.return_value = (visit_date, None)
            response = self.client.post(url)

            if expected_request_status != SAVE_REQUEST_REJECTED:
                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.data['save_request_status'],
                                 expected_request_status)
                self.assertEqual(response.data['save_task_status'],
                                 expected_task_status)

            else:
                self.assertEqual(response.status_code, 403)

    def check_save_request_status(self, mock_scheduler, origin_url,
                                  expected_request_status,
                                  expected_task_status,
                                  scheduler_task_status='next_run_not_scheduled', # noqa
                                  visit_date=None):

        mock_scheduler.get_tasks.return_value = \
            [{
                'priority': 'high',
                'policy': 'oneshot',
                'type': 'origin-update-git',
                'arguments': {
                    'kwargs': {
                        'repo_url': origin_url
                    },
                    'args': []
                },
                'status': scheduler_task_status,
                'id': 1,
             }]

        url = reverse('api-save-origin',
                      url_args={'origin_type': 'git',
                                'origin_url': origin_url})

        with patch('swh.web.common.origin_save._get_visit_info_for_save_request') as mock_visit_date: # noqa
            mock_visit_date.return_value = (visit_date, None)
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            save_request_data = response.data[0]

            self.assertEqual(save_request_data['save_request_status'],
                             expected_request_status)
            self.assertEqual(save_request_data['save_task_status'],
                             expected_task_status)

            # Check that save task status is still available when
            # the scheduler task has been archived
            mock_scheduler.get_tasks.return_value = []
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            save_request_data = response.data[0]
            self.assertEqual(save_request_data['save_task_status'],
                             expected_task_status)

    @patch('swh.web.common.origin_save.scheduler')
    def test_save_request_rejected(self, mock_scheduler):
        origin_url = 'https://github.com/user/illegal_repo'
        self.check_created_save_request_status(mock_scheduler, origin_url,
                                               None, SAVE_REQUEST_REJECTED)
        self.check_save_request_status(mock_scheduler, origin_url,
                                       SAVE_REQUEST_REJECTED,
                                       SAVE_TASK_NOT_CREATED)

    @patch('swh.web.common.origin_save.scheduler')
    def test_save_request_pending(self, mock_scheduler):
        origin_url = 'https://unkwownforge.com/user/repo'
        self.check_created_save_request_status(mock_scheduler, origin_url,
                                               None, SAVE_REQUEST_PENDING,
                                               SAVE_TASK_NOT_CREATED)
        self.check_save_request_status(mock_scheduler, origin_url,
                                       SAVE_REQUEST_PENDING,
                                       SAVE_TASK_NOT_CREATED)

    @patch('swh.web.common.origin_save.scheduler')
    def test_save_request_succeed(self, mock_scheduler):
        origin_url = 'https://github.com/Kitware/CMake'
        self.check_created_save_request_status(mock_scheduler, origin_url,
                                               None, SAVE_REQUEST_ACCEPTED,
                                               SAVE_TASK_NOT_YET_SCHEDULED)
        self.check_save_request_status(mock_scheduler, origin_url,
                                       SAVE_REQUEST_ACCEPTED,
                                       SAVE_TASK_SCHEDULED,
                                       scheduler_task_status='next_run_scheduled') # noqa
        self.check_save_request_status(mock_scheduler, origin_url,
                                       SAVE_REQUEST_ACCEPTED,
                                       SAVE_TASK_SCHEDULED,
                                       scheduler_task_status='completed',
                                       visit_date=None) # noqa
        visit_date = datetime.now(tz=timezone.utc) + timedelta(hours=1)
        self.check_save_request_status(mock_scheduler, origin_url,
                                       SAVE_REQUEST_ACCEPTED,
                                       SAVE_TASK_SUCCEED,
                                       scheduler_task_status='completed',
                                       visit_date=visit_date) # noqa

    @patch('swh.web.common.origin_save.scheduler')
    def test_save_request_failed(self, mock_scheduler):
        origin_url = 'https://gitlab.com/inkscape/inkscape'
        self.check_created_save_request_status(mock_scheduler, origin_url,
                                               None, SAVE_REQUEST_ACCEPTED,
                                               SAVE_TASK_NOT_YET_SCHEDULED)
        self.check_save_request_status(mock_scheduler, origin_url,
                                       SAVE_REQUEST_ACCEPTED,
                                       SAVE_TASK_SCHEDULED,
                                       scheduler_task_status='next_run_scheduled') # noqa
        self.check_save_request_status(mock_scheduler, origin_url,
                                       SAVE_REQUEST_ACCEPTED,
                                       SAVE_TASK_FAILED,
                                       scheduler_task_status='disabled') # noqa

    @patch('swh.web.common.origin_save.scheduler')
    def test_create_save_request_only_when_needed(self, mock_scheduler):
        origin_url = 'https://github.com/webpack/webpack'
        SaveOriginRequest.objects.create(origin_type='git',
                                         origin_url=origin_url,
                                         status=SAVE_REQUEST_ACCEPTED, # noqa
                                         loading_task_id=56)

        self.check_created_save_request_status(mock_scheduler, origin_url,
                                               'next_run_not_scheduled',
                                               SAVE_REQUEST_ACCEPTED,
                                               SAVE_TASK_NOT_YET_SCHEDULED)
        sors = list(SaveOriginRequest.objects.filter(origin_type='git',
                                                     origin_url=origin_url))
        self.assertEqual(len(sors), 1)

        self.check_created_save_request_status(mock_scheduler, origin_url,
                                               'next_run_scheduled',
                                               SAVE_REQUEST_ACCEPTED,
                                               SAVE_TASK_SCHEDULED)
        sors = list(SaveOriginRequest.objects.filter(origin_type='git',
                                                     origin_url=origin_url))
        self.assertEqual(len(sors), 1)

        visit_date = datetime.now(tz=timezone.utc) + timedelta(hours=1)
        self.check_created_save_request_status(mock_scheduler, origin_url,
                                               'completed',
                                               SAVE_REQUEST_ACCEPTED,
                                               SAVE_TASK_NOT_YET_SCHEDULED,
                                               visit_date=visit_date)
        sors = list(SaveOriginRequest.objects.filter(origin_type='git',
                                                     origin_url=origin_url))
        self.assertEqual(len(sors), 2)

        self.check_created_save_request_status(mock_scheduler, origin_url,
                                               'disabled',
                                               SAVE_REQUEST_ACCEPTED,
                                               SAVE_TASK_NOT_YET_SCHEDULED)
        sors = list(SaveOriginRequest.objects.filter(origin_type='git',
                                                     origin_url=origin_url))
        self.assertEqual(len(sors), 3)



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.shortcuts import render

from swh.web.common import service
from swh.web.common.utils import (
    reverse, format_utc_iso_date
)
from swh.web.common.exc import NotFoundExc, handle_view_exception
from swh.web.browse.browseurls import browse_route
from swh.web.browse.utils import (
    gen_person_link, gen_revision_link,
    get_snapshot_context, gen_link,
    gen_snapshot_link, get_swh_persistent_ids
)


@browse_route(r'release/(?P<sha1_git>[0-9a-f]+)/',
              view_name='browse-release')
def release_browse(request, sha1_git):
    """
    Django view that produces an HTML display of a release
    identified by its id.

    The url that points to it is :http:get:`/browse/release/(sha1_git)/`.
    """
    try:
        release = service.lookup_release(sha1_git)
        snapshot_context = None
        origin_info = None
        snapshot_id = request.GET.get('snapshot_id', None)
        origin_type = request.GET.get('origin_type', None)
        origin_url = request.GET.get('origin_url', None)
        if not origin_url:
            origin_url = request.GET.get('origin', None)
        timestamp = request.GET.get('timestamp', None)
        visit_id = request.GET.get('visit_id', None)
        if origin_url:
            try:
                snapshot_context = \
                    get_snapshot_context(snapshot_id, origin_type,
                                         origin_url, timestamp,
                                         visit_id)
            except Exception:
                raw_rel_url = reverse('browse-release',
                                      url_args={'sha1_git': sha1_git})
                error_message = \
                    ('The Software Heritage archive has a release '
                     'with the hash you provided but the origin '
                     'mentioned in your request appears broken: %s. '
                     'Please check the URL and try again.\n\n'
                     'Nevertheless, you can still browse the release '
                     'without origin information: %s'
                        % (gen_link(origin_url), gen_link(raw_rel_url)))

                raise NotFoundExc(error_message)
            origin_info = snapshot_context['origin_info']
        elif snapshot_id:
            snapshot_context = get_snapshot_context(snapshot_id)
    except Exception as exc:
        return handle_view_exception(request, exc)

    release_data = {}

    author_name = 'None'
    release_data['author'] = 'None'
    if release['author']:
        author_name = release['author']['name'] or \
                      release['author']['fullname']
        release_data['author'] = \
            gen_person_link(release['author']['id'], author_name,
                            snapshot_context)
    release_data['date'] = format_utc_iso_date(release['date'])
    release_data['id'] = sha1_git
    release_data['name'] = release['name']
    release_data['synthetic'] = release['synthetic']
    release_data['target type'] = release['target_type']

    if release['target_type'] == 'revision':
        release_data['target'] = \
            gen_revision_link(release['target'],
                              snapshot_context=snapshot_context)
    elif release['target_type'] == 'content':
        content_url = \
            reverse('browse-content',
                    url_args={'query_string': 'sha1_git:' + release['target']})
        release_data['target'] = gen_link(content_url, release['target'])
    elif release['target_type'] == 'directory':
        directory_url = \
            reverse('browse-directory',
                    url_args={'sha1_git': release['target']})
        release_data['target'] = gen_link(directory_url, release['target'])
    elif release['target_type'] == 'release':
        release_url = \
            reverse('browse-release',
                    url_args={'sha1_git': release['target']})
        release_data['target'] = gen_link(release_url, release['target'])

    release_note_lines = []
    if release['message']:
        release_note_lines = release['message'].split('\n')

    vault_cooking = None

    query_params = {}
    if snapshot_id:
        query_params = {'snapshot_id': snapshot_id}
    elif origin_info:
        query_params = {'origin': origin_info['url']}

    target_url = ''
    if release['target_type'] == 'revision':
        target_url = reverse('browse-revision',
                             url_args={'sha1_git': release['target']},
                             query_params=query_params)
        try:
            revision = service.lookup_revision(release['target'])
            vault_cooking = {
                'directory_context': True,
                'directory_id': revision['directory'],
                'revision_context': True,
                'revision_id': release['target']
            }
        except Exception:
            pass
    elif release['target_type'] == 'directory':
        target_url = reverse('browse-directory',
                             url_args={'sha1_git': release['target']},
                             query_params=query_params)
        try:
            revision = service.lookup_directory(release['target'])
            vault_cooking = {
                'directory_context': True,
                'directory_id': revision['directory'],
                'revision_context': False,
                'revision_id': None
            }
        except Exception:
            pass
    elif release['target_type'] == 'content':
        target_url = reverse('browse-content',
                             url_args={'query_string': release['target']},
                             query_params=query_params)
    elif release['target_type'] == 'release':
        target_url = reverse('browse-release',
                             url_args={'sha1_git': release['target']},
                             query_params=query_params)

    release['target_url'] = target_url

    if snapshot_context:
        release_data['snapshot id'] = snapshot_context['snapshot_id']

    if origin_info:
        release_url = reverse('browse-release',
                              url_args={'sha1_git': release['id']})
        release_data['context-independent release'] = \
            gen_link(release_url, link_text='Browse',
                     link_attrs={'class': 'btn btn-default btn-sm',
                                 'role': 'button'})
        release_data['origin id'] = origin_info['id']
        release_data['origin type'] = origin_info['type']
        release_data['origin url'] = gen_link(origin_info['url'],
                                              origin_info['url'])
        browse_snapshot_link = \
            gen_snapshot_link(snapshot_context['snapshot_id'],
                              link_text='Browse',
                              link_attrs={'class': 'btn btn-default btn-sm',
                                          'role': 'button'})
        release_data['snapshot'] = browse_snapshot_link

    swh_objects = [{'type': 'release',
                    'id': sha1_git}]

    if snapshot_context:
        snapshot_id = snapshot_context['snapshot_id']

    if snapshot_id:
        swh_objects.append({'type': 'snapshot',
                            'id': snapshot_id})

    swh_ids = get_swh_persistent_ids(swh_objects, snapshot_context)

    note_header = 'None'
    if len(release_note_lines) > 0:
        note_header = release_note_lines[0]

    release['note_header'] = note_header
    release['note_body'] = '\n'.join(release_note_lines[1:])

    heading = 'Release - %s' % release['name']
    if snapshot_context:
        context_found = 'snapshot: %s' % snapshot_context['snapshot_id']
        if origin_info:
            context_found = 'origin: %s' % origin_info['url']
        heading += ' - %s' % context_found

    return render(request, 'browse/release.html',
                  {'heading': heading,
                   'swh_object_id': swh_ids[0]['swh_id'],
                   'swh_object_name': 'Release',
                   'swh_object_metadata': release_data,
                   'release': release,
                   'snapshot_context': snapshot_context,
                   'show_actions_menu': True,
                   'breadcrumbs': None,
                   'vault_cooking': vault_cooking,
                   'top_right_link': None,
                   'swh_ids': swh_ids})



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.web.common import service
from swh.web.api.apidoc import api_doc
from swh.web.api.apiurls import api_route
from swh.web.api.views.utils import api_lookup


@api_route(r'/person/(?P<person_id>[0-9]+)/', 'api-person')
@api_doc('/person/')
def api_person(request, person_id):
    """
    .. http:get:: /api/1/person/(person_id)/

        Get information about a person in the archive.

        :param int person_id: a person identifier

        :reqheader Accept: the requested response content type,
            either ``application/json`` (default) or ``application/yaml``
        :resheader Content-Type: this depends on :http:header:`Accept` header of request

        :>json string email: the email of the person
        :>json string fullname: the full name of the person: combination of its name and email
        :>json number id: the unique identifier of the person
        :>json string name: the name of the person

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`, :http:method:`options`

        :statuscode 200: no error
        :statuscode 404: requested person can not be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`person/8275/`
    """ # noqa
    return api_lookup(
        service.lookup_person, person_id,
        notfound_msg='Person with id {} not found.'.format(person_id))



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.shortcuts import redirect

from swh.web.common.utils import resolve_swh_persistent_id
from swh.web.common.exc import handle_view_exception


def swh_id_browse(request, swh_id):
    """
    Django view enabling to browse the archive using
    :ref:`persistent-identifiers`.

    The url that points to it is :http:get:`/(swh_id)/`.
    """
    try:
        swh_id_resolved = resolve_swh_persistent_id(
            swh_id, query_params=request.GET)
    except Exception as exc:
        return handle_view_exception(request, exc)

    return redirect(swh_id_resolved['browse_url'])



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.web.common import service
from swh.web.api.apidoc import api_doc
from swh.web.api.apiurls import api_route


@api_route(r'/stat/counters/', 'api-stat-counters')
@api_doc('/stat/counters/', noargs=True)
def api_stats(request):
    """
    .. http:get:: /api/1/stat/counters/

        Get statistics about the content of the archive.

        :>json number content: current number of content objects (aka files) in the archive
        :>json number directory: current number of directory objects in the archive
        :>json number origin: current number of software origins (an origin is a "place" where code
            source can be found, e.g. a git repository, a tarball, ...) in the archive
        :>json number origin_visit: current number of visits on software origins to fill the archive
        :>json number person: current number of persons (code source authors or committers)
            in the archive
        :>json number release: current number of releases objects in the archive
        :>json number revision: current number of revision objects (aka commits) in the archive
        :>json number skipped_content: current number of content objects (aka files) which where
            not inserted in the archive
        :>json number snapshot: current number of snapshot objects (aka set of named branches)
            in the archive

        :reqheader Accept: the requested response content type,
            either ``application/json`` (default) or ``application/yaml``
        :resheader Content-Type: this depends on :http:header:`Accept` header of request

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`, :http:method:`options`

        :statuscode 200: no error

        **Example:**

        .. parsed-literal::

            :swh_web_api:`stat/counters/`
    """ # noqa
    return service.stat_counters()



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


from swh.model.identifiers import (
    CONTENT, DIRECTORY, RELEASE, REVISION, SNAPSHOT
)

from swh.web.common import service
from swh.web.common.utils import resolve_swh_persistent_id
from swh.web.api.apidoc import api_doc
from swh.web.api.apiurls import api_route


@api_route(r'/resolve/(?P<swh_id>.*)/',
           'api-resolve-swh-pid')
@api_doc('/resolve/')
def api_resolve_swh_pid(request, swh_id):
    """
    .. http:get:: /api/1/resolve/(swh_id)/

        Resolve a Software Heritage persistent identifier.

        Try to resolve a provided `persistent identifier <https://docs.softwareheritage.org/devel/swh-model/persistent-identifiers.html>`_
        into an url for browsing the pointed archive object. If the provided
        identifier is valid, the existence of the object in the archive
        will also be checked.

        :param string swh_id: a Software Heritage presistent identifier

        :>json string browse_url: the url for browsing the pointed object
        :>json object metadata: object holding optional parts of the persistent identifier
        :>json string namespace: the persistent identifier namespace
        :>json string object_id: the hash identifier of the pointed object
        :>json string object_type: the type of the pointed object
        :>json number scheme_version: the scheme version of the persistent identifier

        :reqheader Accept: the requested response content type,
            either ``application/json`` (default) or ``application/yaml``
        :resheader Content-Type: this depends on :http:header:`Accept` header of request

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`, :http:method:`options`

        :statuscode 200: no error
        :statuscode 400: an invalid persistent identifier has been provided
        :statuscode 404: the pointed object does not exist in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`resolve/swh:1:rev:96db9023b881d7cd9f379b0c154650d6c108e9a3;origin=https://github.com/openssl/openssl/`
    """  # noqa
    # try to resolve the provided pid
    swh_id_resolved = resolve_swh_persistent_id(swh_id)
    # id is well-formed, now check that the pointed
    # object is present in the archive, NotFoundExc
    # will be raised otherwise
    swh_id_parsed = swh_id_resolved['swh_id_parsed']
    object_type = swh_id_parsed.object_type
    object_id = swh_id_parsed.object_id
    if object_type == CONTENT:
        service.lookup_content('sha1_git:%s' % object_id)
    elif object_type == DIRECTORY:
        service.lookup_directory(object_id)
    elif object_type == RELEASE:
        service.lookup_release(object_id)
    elif object_type == REVISION:
        service.lookup_revision(object_id)
    elif object_type == SNAPSHOT:
        service.lookup_snapshot(object_id)
    # id is well-formed and the pointed object exists
    swh_id_data = swh_id_parsed._asdict()
    swh_id_data['browse_url'] = swh_id_resolved['browse_url']
    return swh_id_data



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


from django.shortcuts import render, redirect
from django.template.defaultfilters import filesizeformat

from swh.web.common import service
from swh.web.common.utils import (
    reverse, gen_path_info
)
from swh.web.common.exc import handle_view_exception, NotFoundExc
from swh.web.browse.utils import (
    get_directory_entries, get_snapshot_context,
    get_readme_to_display, get_swh_persistent_ids,
    gen_link
)

from swh.web.browse.browseurls import browse_route


@browse_route(r'directory/(?P<sha1_git>[0-9a-f]+)/',
              r'directory/(?P<sha1_git>[0-9a-f]+)/(?P<path>.+)/',
              view_name='browse-directory')
def directory_browse(request, sha1_git, path=None):
    """Django view for browsing the content of a directory identified
    by its sha1_git value.

    The url that points to it is :http:get:`/browse/directory/(sha1_git)/[(path)/]`
    """ # noqa
    root_sha1_git = sha1_git
    try:
        if path:
            dir_info = service.lookup_directory_with_path(sha1_git, path)
            # some readme files can reference assets reachable from the
            # browsed directory, handle that special case in order to
            # correctly displayed them
            if dir_info and dir_info['type'] == 'file':
                file_raw_url = reverse(
                    'browse-content-raw',
                    url_args={'query_string': dir_info['checksums']['sha1']})
                return redirect(file_raw_url)
            sha1_git = dir_info['target']

        dirs, files = get_directory_entries(sha1_git)
        origin_type = request.GET.get('origin_type', None)
        origin_url = request.GET.get('origin_url', None)
        if not origin_url:
            origin_url = request.GET.get('origin', None)
        snapshot_context = None
        if origin_url:
            try:
                snapshot_context = get_snapshot_context(None, origin_type,
                                                        origin_url)
            except Exception:
                raw_dir_url = reverse('browse-directory',
                                      url_args={'sha1_git': sha1_git})
                error_message = \
                    ('The Software Heritage archive has a directory '
                     'with the hash you provided but the origin '
                     'mentioned in your request appears broken: %s. '
                     'Please check the URL and try again.\n\n'
                     'Nevertheless, you can still browse the directory '
                     'without origin information: %s'
                        % (gen_link(origin_url), gen_link(raw_dir_url)))

                raise NotFoundExc(error_message)
        if snapshot_context:
            snapshot_context['visit_info'] = None
    except Exception as exc:
        return handle_view_exception(request, exc)

    path_info = gen_path_info(path)

    query_params = {'origin': origin_url}

    breadcrumbs = []
    breadcrumbs.append({'name': root_sha1_git[:7],
                        'url': reverse('browse-directory',
                                       url_args={'sha1_git': root_sha1_git},
                                       query_params=query_params)})
    for pi in path_info:
        breadcrumbs.append({'name': pi['name'],
                            'url': reverse('browse-directory',
                                           url_args={'sha1_git': root_sha1_git,
                                                     'path': pi['path']},
                                           query_params=query_params)})

    path = '' if path is None else (path + '/')

    for d in dirs:
        if d['type'] == 'rev':
            d['url'] = reverse('browse-revision',
                               url_args={'sha1_git': d['target']},
                               query_params=query_params)
        else:
            d['url'] = reverse('browse-directory',
                               url_args={'sha1_git': root_sha1_git,
                                         'path': path + d['name']},
                               query_params=query_params)

    sum_file_sizes = 0

    readmes = {}

    for f in files:
        query_string = 'sha1_git:' + f['target']
        f['url'] = reverse('browse-content',
                           url_args={'query_string': query_string},
                           query_params={'path': root_sha1_git + '/' +
                                         path + f['name'],
                                         'origin': origin_url})
        if f['length'] is not None:
            sum_file_sizes += f['length']
            f['length'] = filesizeformat(f['length'])
        if f['name'].lower().startswith('readme'):
            readmes[f['name']] = f['checksums']['sha1']

    readme_name, readme_url, readme_html = get_readme_to_display(readmes)

    sum_file_sizes = filesizeformat(sum_file_sizes)

    dir_metadata = {'id': sha1_git,
                    'number of regular files': len(files),
                    'number of subdirectories': len(dirs),
                    'sum of regular file sizes': sum_file_sizes}

    vault_cooking = {
        'directory_context': True,
        'directory_id': sha1_git,
        'revision_context': False,
        'revision_id': None
    }

    swh_ids = get_swh_persistent_ids([{'type': 'directory',
                                       'id': sha1_git}])

    heading = 'Directory - %s' % sha1_git
    if breadcrumbs:
        dir_path = '/'.join([bc['name'] for bc in breadcrumbs]) + '/'
        heading += ' - %s' % dir_path

    return render(request, 'browse/directory.html',
                  {'heading': heading,
                   'swh_object_id': swh_ids[0]['swh_id'],
                   'swh_object_name': 'Directory',
                   'swh_object_metadata': dir_metadata,
                   'dirs': dirs,
                   'files': files,
                   'breadcrumbs': breadcrumbs,
                   'top_right_link': None,
                   'readme_name': readme_name,
                   'readme_url': readme_url,
                   'readme_html': readme_html,
                   'snapshot_context': snapshot_context,
                   'vault_cooking': vault_cooking,
                   'show_actions_menu': True,
                   'swh_ids': swh_ids})



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os

from django.core.wsgi import get_wsgi_application

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "swh.web.settings.production")

application = get_wsgi_application()



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json

from distutils.util import strtobool

from django.core.cache import caches
from django.http import HttpResponse
from django.shortcuts import render, redirect
from django.views.decorators.cache import never_cache

from swh.web.common import service
from swh.web.common.origin_visits import get_origin_visits
from swh.web.common.utils import (
    reverse, format_utc_iso_date, parse_timestamp
)
from swh.web.common.exc import handle_view_exception
from swh.web.browse.utils import (
    get_origin_info, get_snapshot_context
)
from swh.web.browse.browseurls import browse_route
from swh.web.misc.coverage import code_providers

from .utils.snapshot_context import (
    browse_snapshot_directory, browse_snapshot_content,
    browse_snapshot_log, browse_snapshot_branches,
    browse_snapshot_releases
)

@browse_route(r'origin/(?P<origin_type>[a-z]+)/url/(?P<origin_url>.+)/visit/(?P<timestamp>.+)/directory/', # noqa
              r'origin/(?P<origin_type>[a-z]+)/url/(?P<origin_url>.+)/visit/(?P<timestamp>.+)/directory/(?P<path>.+)/', # noqa
              r'origin/(?P<origin_type>[a-z]+)/url/(?P<origin_url>.+)/directory/', # noqa
              r'origin/(?P<origin_type>[a-z]+)/url/(?P<origin_url>.+)/directory/(?P<path>.+)/', # noqa
              r'origin/(?P<origin_url>.+)/visit/(?P<timestamp>.+)/directory/', # noqa
              r'origin/(?P<origin_url>.+)/visit/(?P<timestamp>.+)/directory/(?P<path>.+)/', # noqa
              r'origin/(?P<origin_url>.+)/directory/', # noqa
              r'origin/(?P<origin_url>.+)/directory/(?P<path>.+)/', # noqa
              view_name='browse-origin-directory')
def origin_directory_browse(request, origin_url, origin_type=None,
                            timestamp=None, path=None):
    """Django view for browsing the content of a directory associated
    to an origin for a given visit.

    The url scheme that points to it is the following:

        * :http:get:`/browse/origin/[(origin_type)/url/](origin_url)/directory/[(path)/]`
        * :http:get:`/browse/origin/[(origin_type)/url/](origin_url)/visit/(timestamp)/directory/[(path)/]`
    """ # noqa
    return browse_snapshot_directory(
        request, origin_type=origin_type, origin_url=origin_url,
        timestamp=timestamp, path=path)


@browse_route(r'origin/(?P<origin_type>[a-z]+)/url/(?P<origin_url>.+)/visit/(?P<timestamp>.+)/content/(?P<path>.+)/', # noqa
              r'origin/(?P<origin_type>[a-z]+)/url/(?P<origin_url>.+)/content/(?P<path>.+)/', # noqa
              r'origin/(?P<origin_url>.+)/visit/(?P<timestamp>.+)/content/(?P<path>.+)/', # noqa
              r'origin/(?P<origin_url>.+)/content/(?P<path>.+)/', # noqa
              view_name='browse-origin-content')
def origin_content_browse(request, origin_url, origin_type=None, path=None,
                          timestamp=None):
    """Django view that produces an HTML display of a content
    associated to an origin for a given visit.

    The url scheme that points to it is the following:

        * :http:get:`/browse/origin/[(origin_type)/url/](origin_url)/content/(path)/`
        * :http:get:`/browse/origin/[(origin_type)/url/](origin_url)/visit/(timestamp)/content/(path)/`

    """ # noqa
    return browse_snapshot_content(request, origin_type=origin_type,
                                   origin_url=origin_url, timestamp=timestamp,
                                   path=path)


PER_PAGE = 20


@browse_route(r'origin/(?P<origin_type>[a-z]+)/url/(?P<origin_url>.+)/visit/(?P<timestamp>.+)/log/', # noqa
              r'origin/(?P<origin_type>[a-z]+)/url/(?P<origin_url>.+)/log/',
              r'origin/(?P<origin_url>.+)/visit/(?P<timestamp>.+)/log/', # noqa
              r'origin/(?P<origin_url>.+)/log/',
              view_name='browse-origin-log')
def origin_log_browse(request, origin_url, origin_type=None, timestamp=None):
    """Django view that produces an HTML display of revisions history (aka
    the commit log) associated to a software origin.

    The url scheme that points to it is the following:

        * :http:get:`/browse/origin/[(origin_type)/url/](origin_url)/log/`
        * :http:get:`/browse/origin/[(origin_type)/url/](origin_url)/visit/(timestamp)/log/`
    """ # noqa
    return browse_snapshot_log(request, origin_type=origin_type,
                               origin_url=origin_url, timestamp=timestamp)

@browse_route(r'origin/(?P<origin_type>[a-z]+)/url/(?P<origin_url>.+)/visit/(?P<timestamp>.+)/branches/', # noqa
              r'origin/(?P<origin_type>[a-z]+)/url/(?P<origin_url>.+)/branches/', # noqa
              r'origin/(?P<origin_url>.+)/visit/(?P<timestamp>.+)/branches/', # noqa
              r'origin/(?P<origin_url>.+)/branches/', # noqa
              view_name='browse-origin-branches')
def origin_branches_browse(request, origin_url, origin_type=None,
                           timestamp=None):
    """Django view that produces an HTML display of the list of branches
    associated to an origin for a given visit.

    The url scheme that points to it is the following:

        * :http:get:`/browse/origin/[(origin_type)/url/](origin_url)/branches/`
        * :http:get:`/browse/origin/[(origin_type)/url/](origin_url)/visit/(timestamp)/branches/`

    """ # noqa
    return browse_snapshot_branches(request, origin_type=origin_type,
                                    origin_url=origin_url, timestamp=timestamp)


@browse_route(r'origin/(?P<origin_type>[a-z]+)/url/(?P<origin_url>.+)/visit/(?P<timestamp>.+)/releases/', # noqa
              r'origin/(?P<origin_type>[a-z]+)/url/(?P<origin_url>.+)/releases/', # noqa
              r'origin/(?P<origin_url>.+)/visit/(?P<timestamp>.+)/releases/', # noqa
              r'origin/(?P<origin_url>.+)/releases/', # noqa
              view_name='browse-origin-releases')
def origin_releases_browse(request, origin_url, origin_type=None,
                           timestamp=None):
    """Django view that produces an HTML display of the list of releases
    associated to an origin for a given visit.

    The url scheme that points to it is the following:

        * :http:get:`/browse/origin/[(origin_type)/url/](origin_url)/releases/`
        * :http:get:`/browse/origin/[(origin_type)/url/](origin_url)/visit/(timestamp)/releases/`

    """ # noqa
    return browse_snapshot_releases(request, origin_type=origin_type,
                                    origin_url=origin_url, timestamp=timestamp)


@browse_route(r'origin/(?P<origin_type>[a-z]+)/url/(?P<origin_url>.+)/visits/',
              r'origin/(?P<origin_url>.+)/visits/',
              view_name='browse-origin-visits')
def origin_visits_browse(request, origin_url, origin_type=None):
    """Django view that produces an HTML display of visits reporting
    for a swh origin identified by its id or its url.

    The url that points to it is :http:get:`/browse/origin/[(origin_type)/url/](origin_url)/visits/`.
    """ # noqa
    try:
        origin_info = get_origin_info(origin_url, origin_type)
        origin_visits = get_origin_visits(origin_info)
        snapshot_context = get_snapshot_context(origin_type=origin_type,
                                                origin_url=origin_url)
    except Exception as exc:
        return handle_view_exception(request, exc)

    for i, visit in enumerate(origin_visits):
        url_date = format_utc_iso_date(visit['date'], '%Y-%m-%dT%H:%M:%SZ')
        visit['fmt_date'] = format_utc_iso_date(visit['date'])
        query_params = {}
        if i < len(origin_visits) - 1:
            if visit['date'] == origin_visits[i+1]['date']:
                query_params = {'visit_id': visit['visit']}
        if i > 0:
            if visit['date'] == origin_visits[i-1]['date']:
                query_params = {'visit_id': visit['visit']}

        snapshot = visit['snapshot'] if visit['snapshot'] else ''

        visit['browse_url'] = reverse('browse-origin-directory',
                                      url_args={'origin_type': origin_type,
                                                'origin_url': origin_url,
                                                'timestamp': url_date},
                                      query_params=query_params)
        if not snapshot:
            visit['snapshot'] = ''
        visit['date'] = parse_timestamp(visit['date']).timestamp()

    heading = 'Origin visits - %s' % origin_url

    return render(request, 'browse/origin-visits.html',
                  {'heading': heading,
                   'swh_object_name': 'Visits',
                   'swh_object_metadata': origin_info,
                   'origin_visits': origin_visits,
                   'origin_info': origin_info,
                   'snapshot_context': snapshot_context,
                   'vault_cooking': None,
                   'show_actions_menu': False})


@browse_route(r'origin/search/(?P<url_pattern>.+)/',
              view_name='browse-origin-search')
def _origin_search(request, url_pattern):
    """Internal browse endpoint to search for origins whose urls contain
    a provided string pattern or match a provided regular expression.
    The search is performed in a case insensitive way.
    """
    offset = int(request.GET.get('offset', '0'))
    limit = int(request.GET.get('limit', '50'))
    regexp = request.GET.get('regexp', 'false')
    with_visit = request.GET.get('with_visit', 'false')

    url_pattern = url_pattern.replace('///', '\\')

    try:
        results = service.search_origin(url_pattern, offset, limit,
                                        bool(strtobool(regexp)),
                                        bool(strtobool(with_visit)))

        results = json.dumps(list(results), sort_keys=True, indent=4,
                             separators=(',', ': '))
    except Exception as exc:
        return handle_view_exception(request, exc, html_response=False)

    return HttpResponse(results, content_type='application/json')


@browse_route(r'origin/coverage_count/',
              view_name='browse-origin-coverage-count')
@never_cache
def _origin_coverage_count(request):
    """Internal browse endpoint to count the number of origins associated
    to each code provider declared in the archive coverage list.
    As this operation takes some times, we execute it once per day and
    cache its results to database. The cached origin counts are then served.
    Cache management is handled in the implementation to avoid sending
    the same count query twice to the storage database.
    """
    try:
        cache = caches['db_cache']
        results = []
        for code_provider in code_providers:
            provider_id = code_provider['provider_id']
            url_regexp = code_provider['origin_url_regexp']
            cache_key = '%s_origins_count' % provider_id
            prev_cache_key = '%s_origins_prev_count' % provider_id
            # get cached origin count
            origin_count = cache.get(cache_key, -2)
            # cache entry has expired or does not exist
            if origin_count == -2:
                # mark the origin count as processing
                cache.set(cache_key, -1, timeout=10*60)
                # execute long count query
                origin_count = service.storage.origin_count(url_regexp,
                                                            regexp=True)
                # cache count result
                cache.set(cache_key, origin_count, timeout=24*60*60)
                cache.set(prev_cache_key, origin_count, timeout=None)
            # origin count is currently processing
            elif origin_count == -1:
                # return previous count if it exists
                origin_count = cache.get(prev_cache_key, -1)
            results.append({
                'provider_id': provider_id,
                'origin_count': origin_count,
                'origin_types': code_provider['origin_types']
            })
        results = json.dumps(results)
    except Exception as exc:
        return handle_view_exception(request, exc, html_response=False)

    return HttpResponse(results, content_type='application/json')


@browse_route(r'origin/(?P<origin_id>[0-9]+)/latest_snapshot/',
              view_name='browse-origin-latest-snapshot')
def _origin_latest_snapshot(request, origin_id):
    """
    Internal browse endpoint used to check if an origin has already
    been visited by Software Heritage and has at least one full visit.
    """
    result = \
        service.lookup_latest_origin_snapshot(origin_id,
                                              allowed_statuses=['full',
                                                                'partial'])

    result = json.dumps(result, sort_keys=True, indent=4,
                        separators=(',', ': '))

    return HttpResponse(result, content_type='application/json')


@browse_route(r'origin/(?P<origin_type>[a-z]+)/url/(?P<origin_url>.+)/',
              r'origin/(?P<origin_url>.+)/',
              view_name='browse-origin')
def origin_browse(request, origin_url, origin_type=None):
    """Django view that redirects to the display of the latest archived
    snapshot for a given software origin.
    """ # noqa
    last_snapshot_url = reverse('browse-origin-directory',
                                url_args={'origin_type': origin_type,
                                          'origin_url': origin_url})
    return redirect(last_snapshot_url)



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from unittest.mock import patch

from django.utils.html import escape

from hypothesis import given

from swh.web.browse.utils import (
    get_mimetype_and_encoding_for_content, prepare_content_for_display,
    _reencode_content
)
from swh.web.common.exc import NotFoundExc
from swh.web.common.utils import reverse, get_swh_persistent_id
from swh.web.common.utils import gen_path_info
from swh.web.tests.strategies import (
    content, content_text_non_utf8, content_text_no_highlight,
    content_image_type, content_text, invalid_sha1, unknown_content
)
from swh.web.tests.testcase import WebTestCase


class SwhBrowseContentTest(WebTestCase):

    @given(content())
    def test_content_view_text(self, content):

        sha1_git = content['sha1_git']

        url = reverse('browse-content',
                      url_args={'query_string': content['sha1']},
                      query_params={'path': content['path']})

        url_raw = reverse('browse-content-raw',
                          url_args={'query_string': content['sha1']})

        resp = self.client.get(url)

        content_display = self._process_content_for_display(content)
        mimetype = content_display['mimetype']

        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed('browse/content.html')

        if mimetype.startswith('text/'):
            self.assertContains(resp, '<code class="%s">' %
                                      content_display['language'])
            self.assertContains(resp, escape(content_display['content_data']))
        self.assertContains(resp, url_raw)

        swh_cnt_id = get_swh_persistent_id('content', sha1_git)
        swh_cnt_id_url = reverse('browse-swh-id',
                                 url_args={'swh_id': swh_cnt_id})
        self.assertContains(resp, swh_cnt_id)
        self.assertContains(resp, swh_cnt_id_url)

    @given(content_text_no_highlight())
    def test_content_view_text_no_highlight(self, content):

        sha1_git = content['sha1_git']

        url = reverse('browse-content',
                      url_args={'query_string': content['sha1']})

        url_raw = reverse('browse-content-raw',
                          url_args={'query_string': content['sha1']})

        resp = self.client.get(url)

        content_display = self._process_content_for_display(content)

        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed('browse/content.html')

        self.assertContains(resp, '<code class="nohighlight">')
        self.assertContains(resp, escape(content_display['content_data'])) # noqa
        self.assertContains(resp, url_raw)

        swh_cnt_id = get_swh_persistent_id('content', sha1_git)
        swh_cnt_id_url = reverse('browse-swh-id',
                                 url_args={'swh_id': swh_cnt_id})

        self.assertContains(resp, swh_cnt_id)
        self.assertContains(resp, swh_cnt_id_url)

    @given(content_text_non_utf8())
    def test_content_view_no_utf8_text(self, content):

        sha1_git = content['sha1_git']

        url = reverse('browse-content',
                      url_args={'query_string': content['sha1']})

        resp = self.client.get(url)

        content_display = self._process_content_for_display(content)

        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed('browse/content.html')
        swh_cnt_id = get_swh_persistent_id('content', sha1_git)
        swh_cnt_id_url = reverse('browse-swh-id',
                                 url_args={'swh_id': swh_cnt_id})
        self.assertContains(resp, swh_cnt_id_url)
        self.assertContains(resp, escape(content_display['content_data']))

    @given(content_image_type())
    def test_content_view_image(self, content):

        url = reverse('browse-content',
                      url_args={'query_string': content['sha1']})

        url_raw = reverse('browse-content-raw',
                          url_args={'query_string': content['sha1']})

        resp = self.client.get(url)

        content_display = self._process_content_for_display(content)
        mimetype = content_display['mimetype']
        content_data = content_display['content_data']

        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed('browse/content.html')

        self.assertContains(resp, '<img src="data:%s;base64,%s"/>'
                                  % (mimetype, content_data.decode('utf-8')))
        self.assertContains(resp, url_raw)

    @given(content())
    def test_content_view_with_path(self, content):

        path = content['path']

        url = reverse('browse-content',
                      url_args={'query_string': content['sha1']},
                      query_params={'path': path})

        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed('browse/content.html')

        self.assertContains(resp, '<nav class="bread-crumbs')

        content_display = self._process_content_for_display(content)
        mimetype = content_display['mimetype']

        if mimetype.startswith('text/'):
            hljs_language = content['hljs_language']
            self.assertContains(resp, '<code class="%s">' % hljs_language)
            self.assertContains(resp, escape(content_display['content_data']))

        split_path = path.split('/')

        root_dir_sha1 = split_path[0]
        filename = split_path[-1]
        path = path.replace(root_dir_sha1 + '/', '').replace(filename, '')

        path_info = gen_path_info(path)

        root_dir_url = reverse('browse-directory',
                               url_args={'sha1_git': root_dir_sha1})

        self.assertContains(resp, '<li class="swh-path">',
                            count=len(path_info)+1)

        self.assertContains(resp, '<a href="' + root_dir_url + '">' +
                            root_dir_sha1[:7] + '</a>')

        for p in path_info:
            dir_url = reverse('browse-directory',
                              url_args={'sha1_git': root_dir_sha1,
                                        'path': p['path']})
            self.assertContains(resp, '<a href="' + dir_url + '">' +
                                p['name'] + '</a>')

        self.assertContains(resp, '<li>' + filename + '</li>')

        url_raw = reverse('browse-content-raw',
                          url_args={'query_string': content['sha1']},
                          query_params={'filename': filename})
        self.assertContains(resp, url_raw)

        url = reverse('browse-content',
                      url_args={'query_string': content['sha1']},
                      query_params={'path': filename})

        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed('browse/content.html')

        self.assertNotContains(resp, '<nav class="bread-crumbs')

    @given(content_text())
    def test_content_raw_text(self, content):

        url = reverse('browse-content-raw',
                      url_args={'query_string': content['sha1']})

        resp = self.client.get(url)

        content_data = self.content_get(content['sha1'])['data']

        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp['Content-Type'], 'text/plain')
        self.assertEqual(resp['Content-disposition'],
                         'filename=%s_%s' % ('sha1', content['sha1']))
        self.assertEqual(resp.content, content_data)

        filename = content['path'].split('/')[-1]

        url = reverse('browse-content-raw',
                      url_args={'query_string': content['sha1']}, # noqa
                      query_params={'filename': filename})

        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp['Content-Type'], 'text/plain')
        self.assertEqual(resp['Content-disposition'],
                         'filename=%s' % filename)
        self.assertEqual(resp.content, content_data)

    @given(content_text_non_utf8())
    def test_content_raw_no_utf8_text(self, content):

        url = reverse('browse-content-raw',
                      url_args={'query_string': content['sha1']})

        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        _, encoding = get_mimetype_and_encoding_for_content(resp.content)
        self.assertEqual(encoding, content['encoding'])

    @given(content_image_type())
    def test_content_raw_bin(self, content):

        url = reverse('browse-content-raw',
                      url_args={'query_string': content['sha1']})

        resp = self.client.get(url)

        filename = content['path'].split('/')[-1]
        content_data = self.content_get(content['sha1'])['data']

        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp['Content-Type'], 'application/octet-stream')
        self.assertEqual(resp['Content-disposition'],
                         'attachment; filename=%s_%s' %
                         ('sha1', content['sha1']))
        self.assertEqual(resp.content, content_data)

        url = reverse('browse-content-raw',
                      url_args={'query_string': content['sha1']},
                      query_params={'filename': filename})

        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp['Content-Type'], 'application/octet-stream')
        self.assertEqual(resp['Content-disposition'],
                         'attachment; filename=%s' % filename)
        self.assertEqual(resp.content, content_data)

    @given(invalid_sha1(), unknown_content())
    def test_content_request_errors(self, invalid_sha1, unknown_content):

        url = reverse('browse-content',
                      url_args={'query_string': invalid_sha1})
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 400)
        self.assertTemplateUsed('error.html')

        url = reverse('browse-content',
                      url_args={'query_string': unknown_content['sha1']})
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 404)
        self.assertTemplateUsed('error.html')

    @patch('swh.web.browse.utils.service')
    @given(content())
    def test_content_bytes_missing(self, mock_service, content):

        content_data = self.content_get_metadata(content['sha1'])
        content_data['data'] = None

        mock_service.lookup_content.return_value = content_data
        mock_service.lookup_content_raw.side_effect = NotFoundExc(
            'Content bytes not available!')

        url = reverse('browse-content',
                      url_args={'query_string': content['sha1']})

        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 404)
        self.assertTemplateUsed('browse/content.html')

    @patch('swh.web.browse.views.content.request_content')
    def test_content_too_large(self, mock_request_content):
        stub_content_too_large_data = {
            'checksums': {
                'sha1': '8624bcdae55baeef00cd11d5dfcfa60f68710a02',
                'sha1_git': '94a9ed024d3859793618152ea559a168bbcbb5e2',
                'sha256': ('8ceb4b9ee5adedde47b31e975c1d90c73ad27b6b16'
                           '5a1dcd80c7c545eb65b903'),
                'blake2s256': ('38702b7168c7785bfe748b51b45d9856070ba90'
                               'f9dc6d90f2ea75d4356411ffe')
            },
            'length': 3000000,
            'raw_data': None,
            'mimetype': 'text/plain',
            'encoding': 'us-ascii',
            'language': 'not detected',
            'licenses': 'GPL',
            'error_code': 200,
            'error_message': '',
            'error_description': ''
        }

        content_sha1 = stub_content_too_large_data['checksums']['sha1']

        mock_request_content.return_value = stub_content_too_large_data

        url = reverse('browse-content',
                      url_args={'query_string': content_sha1})

        url_raw = reverse('browse-content-raw',
                          url_args={'query_string': content_sha1})

        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed('browse/content.html')

        self.assertContains(resp, 'Content is too large to be displayed')
        self.assertContains(resp, url_raw)

    def _process_content_for_display(self, content):
        content_data = self.content_get(content['sha1'])

        mime_type, encoding = get_mimetype_and_encoding_for_content(
            content_data['data'])

        mime_type, content_data = _reencode_content(mime_type, encoding,
                                                    content_data['data'])

        return prepare_content_for_display(content_data, mime_type,
                                           content['path'])



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


from django.shortcuts import render
from swh.web.common import service
from swh.web.common.exc import handle_view_exception
from swh.web.browse.browseurls import browse_route
from swh.web.browse.utils import get_snapshot_context


@browse_route(r'person/(?P<person_id>[0-9]+)/',
              view_name='browse-person')
def person_browse(request, person_id):
    """
    Django view that produces an HTML display of a swh person
    identified by its id.

    The url that points to it is :http:get:`/browse/person/(person_id)/`.
    """
    try:
        snapshot_context = None
        origin_type = request.GET.get('origin_type', None)
        origin_url = request.GET.get('origin_url', None)
        if not origin_url:
            origin_url = request.GET.get('origin', None)
        snapshot_id = request.GET.get('snapshot_id', None)
        if origin_url:
            snapshot_context = get_snapshot_context(None, origin_type,
                                                    origin_url)
        elif snapshot_id:
            snapshot_context = get_snapshot_context(snapshot_id)
        person = service.lookup_person(person_id)
    except Exception as exc:
        return handle_view_exception(request, exc)

    heading = 'Person - %s' % person['fullname']
    if snapshot_context:
        context_found = 'snapshot: %s' % snapshot_context['snapshot_id']
        if origin_url:
            context_found = 'origin: %s' % origin_url
        heading += ' - %s' % context_found

    return render(request, 'browse/person.html',
                  {'heading': heading,
                   'swh_object_name': 'Person',
                   'swh_object_metadata': person,
                   'snapshot_context': snapshot_context,
                   'vault_cooking': None,
                   'show_actions_menu': False})



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import docutils.parsers.rst
import docutils.utils
import re
import requests

from datetime import datetime, timezone
from dateutil import parser as date_parser
from dateutil import tz

from django.urls import reverse as django_reverse
from django.http import QueryDict

from swh.model.exceptions import ValidationError
from swh.model.identifiers import (
    persistent_identifier, parse_persistent_identifier,
    CONTENT, DIRECTORY, RELEASE, REVISION, SNAPSHOT
)

from swh.web.common.exc import BadInputExc
from swh.web.config import get_config

swh_object_icons = {
    'branch': 'fa fa-code-fork',
    'branches': 'fa fa-code-fork',
    'content': 'fa fa-file-text',
    'directory': 'fa fa-folder',
    'person': 'fa fa-user',
    'revisions history': 'fa fa-history',
    'release': 'fa fa-tag',
    'releases': 'fa fa-tag',
    'revision': 'octicon-git-commit',
    'snapshot': 'fa fa-camera',
    'visits': 'fa fa-calendar',
}


def reverse(viewname, url_args=None, query_params=None,
            current_app=None, urlconf=None):
    """An override of django reverse function supporting query parameters.

    Args:
        viewname (str): the name of the django view from which to compute a url
        url_args (dict): dictionary of url arguments indexed by their names
        query_params (dict): dictionary of query parameters to append to the
            reversed url
        current_app (str): the name of the django app tighten to the view
        urlconf (str): url configuration module

    Returns:
        str: the url of the requested view with processed arguments and
        query parameters
    """

    if url_args:
        url_args = {k: v for k, v in url_args.items() if v is not None}

    url = django_reverse(viewname, urlconf=urlconf, kwargs=url_args,
                         current_app=current_app)

    if query_params:
        query_params = {k: v for k, v in query_params.items() if v}

    if query_params and len(query_params) > 0:
        query_dict = QueryDict('', mutable=True)
        for k in sorted(query_params.keys()):
            query_dict[k] = query_params[k]
        url += ('?' + query_dict.urlencode(safe='/;:'))

    return url


def datetime_to_utc(date):
    """Returns datetime in UTC without timezone info

    Args:
        date (datetime.datetime): input datetime with timezone info

    Returns:
        datetime.datetime: datetime in UTC without timezone info
    """
    if date.tzinfo:
        return date.astimezone(tz.gettz('UTC')).replace(tzinfo=timezone.utc)
    else:
        return date


def parse_timestamp(timestamp):
    """Given a time or timestamp (as string), parse the result as UTC datetime.

    Returns:
        datetime.datetime: a timezone-aware datetime representing the
            parsed value or None if the parsing fails.

    Samples:
        - 2016-01-12
        - 2016-01-12T09:19:12+0100
        - Today is January 1, 2047 at 8:21:00AM
        - 1452591542

    """
    if not timestamp:
        return None

    try:
        date = date_parser.parse(timestamp, ignoretz=False, fuzzy=True)
        return datetime_to_utc(date)
    except Exception:
        try:
            return datetime.utcfromtimestamp(float(timestamp)).replace(
                tzinfo=timezone.utc)
        except (ValueError, OverflowError) as e:
            raise BadInputExc(e)


def shorten_path(path):
    """Shorten the given path: for each hash present, only return the first
    8 characters followed by an ellipsis"""

    sha256_re = r'([0-9a-f]{8})[0-9a-z]{56}'
    sha1_re = r'([0-9a-f]{8})[0-9a-f]{32}'

    ret = re.sub(sha256_re, r'\1...', path)
    return re.sub(sha1_re, r'\1...', ret)


def format_utc_iso_date(iso_date, fmt='%d %B %Y, %H:%M UTC'):
    """Turns a string representation of an ISO 8601 date string
    to UTC and format it into a more human readable one.

    For instance, from the following input
    string: '2017-05-04T13:27:13+02:00' the following one
    is returned: '04 May 2017, 11:27 UTC'.
    Custom format string may also be provided
    as parameter

    Args:
        iso_date (str): a string representation of an ISO 8601 date
        fmt (str): optional date formatting string

    Returns:
        str: a formatted string representation of the input iso date
    """
    if not iso_date:
        return iso_date
    date = parse_timestamp(iso_date)
    return date.strftime(fmt)


def gen_path_info(path):
    """Function to generate path data navigation for use
    with a breadcrumb in the swh web ui.

    For instance, from a path /folder1/folder2/folder3,
    it returns the following list::

        [{'name': 'folder1', 'path': 'folder1'},
         {'name': 'folder2', 'path': 'folder1/folder2'},
         {'name': 'folder3', 'path': 'folder1/folder2/folder3'}]

    Args:
        path: a filesystem path

    Returns:
        list: a list of path data for navigation as illustrated above.

    """
    path_info = []
    if path:
        sub_paths = path.strip('/').split('/')
        path_from_root = ''
        for p in sub_paths:
            path_from_root += '/' + p
            path_info.append({'name': p,
                              'path': path_from_root.strip('/')})
    return path_info


def get_swh_persistent_id(object_type, object_id, scheme_version=1):
    """
    Returns the persistent identifier for a swh object based on:

        * the object type
        * the object id
        * the swh identifiers scheme version

    Args:
        object_type (str): the swh object type
            (content/directory/release/revision/snapshot)
        object_id (str): the swh object id (hexadecimal representation
            of its hash value)
        scheme_version (int): the scheme version of the swh
            persistent identifiers

    Returns:
        str: the swh object persistent identifier

    Raises:
        BadInputExc: if the provided parameters do not enable to
            generate a valid identifier
    """
    try:
        swh_id = persistent_identifier(object_type, object_id, scheme_version)
    except ValidationError as e:
        raise BadInputExc('Invalid object (%s) for swh persistent id. %s' %
                          (object_id, e))
    else:
        return swh_id


def resolve_swh_persistent_id(swh_id, query_params=None):
    """
    Try to resolve a Software Heritage persistent id into an url for
    browsing the pointed object.

    Args:
        swh_id (str): a Software Heritage persistent identifier
        query_params (django.http.QueryDict): optional dict filled with
            query parameters to append to the browse url

    Returns:
        dict: a dict with the following keys:

            * **swh_id_parsed (swh.model.identifiers.PersistentId)**: the parsed identifier
            * **browse_url (str)**: the url for browsing the pointed object

    Raises:
        BadInputExc: if the provided identifier can not be parsed
    """ # noqa
    try:
        swh_id_parsed = parse_persistent_identifier(swh_id)
        object_type = swh_id_parsed.object_type
        object_id = swh_id_parsed.object_id
        browse_url = None
        query_dict = QueryDict('', mutable=True)
        if query_params and len(query_params) > 0:
            for k in sorted(query_params.keys()):
                query_dict[k] = query_params[k]
        if 'origin' in swh_id_parsed.metadata:
            query_dict['origin'] = swh_id_parsed.metadata['origin']
        if object_type == CONTENT:
            query_string = 'sha1_git:' + object_id
            fragment = ''
            if 'lines' in swh_id_parsed.metadata:
                lines = swh_id_parsed.metadata['lines'].split('-')
                fragment += '#L' + lines[0]
                if len(lines) > 1:
                    fragment += '-L' + lines[1]
            browse_url = reverse('browse-content',
                                 url_args={'query_string': query_string},
                                 query_params=query_dict) + fragment
        elif object_type == DIRECTORY:
            browse_url = reverse('browse-directory',
                                 url_args={'sha1_git': object_id},
                                 query_params=query_dict)
        elif object_type == RELEASE:
            browse_url = reverse('browse-release',
                                 url_args={'sha1_git': object_id},
                                 query_params=query_dict)
        elif object_type == REVISION:
            browse_url = reverse('browse-revision',
                                 url_args={'sha1_git': object_id},
                                 query_params=query_dict)
        elif object_type == SNAPSHOT:
            browse_url = reverse('browse-snapshot',
                                 url_args={'snapshot_id': object_id},
                                 query_params=query_dict)
    except ValidationError as ve:
        raise BadInputExc('Error when parsing identifier. %s' %
                          ' '.join(ve.messages))
    else:
        return {'swh_id_parsed': swh_id_parsed,
                'browse_url': browse_url}


def parse_rst(text, report_level=2):
    """
    Parse a reStructuredText string with docutils.

    Args:
        text (str): string with reStructuredText markups in it
        report_level (int): level of docutils report messages to print
            (1 info 2 warning 3 error 4 severe 5 none)

    Returns:
        docutils.nodes.document: a parsed docutils document
    """
    parser = docutils.parsers.rst.Parser()
    components = (docutils.parsers.rst.Parser,)
    settings = docutils.frontend.OptionParser(
        components=components).get_default_values()
    settings.report_level = report_level
    document = docutils.utils.new_document('rst-doc', settings=settings)
    parser.parse(text, document)
    return document


def get_client_ip(request):
    """
    Return the client IP address from an incoming HTTP request.

    Args:
        request (django.http.HttpRequest): the incoming HTTP request

    Returns:
        str: The client IP address
    """
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


def is_recaptcha_valid(request, recaptcha_response):
    """
    Verify if the response for Google reCAPTCHA is valid.

    Args:
        request (django.http.HttpRequest): the incoming HTTP request
        recaptcha_response (str): the reCAPTCHA response

    Returns:
        bool: Whether the reCAPTCHA response is valid or not
    """
    config = get_config()
    if config['grecaptcha']['activated'] is False:
        recaptcha_valid = True
    else:
        recaptcha_valid = requests.post(
            config['grecaptcha']['validation_url'],
            data={
                'secret': config['grecaptcha']['private_key'],
                'response': recaptcha_response,
                'remoteip': get_client_ip(request)
            },
            verify=True
        ).json().get("success", False)
    return recaptcha_valid


def context_processor(request):
    """
    Django context processor used to inject variables
    in all swh-web templates.
    """
    config = get_config()
    return {'swh_object_icons': swh_object_icons,
            'grecaptcha_activated': config['grecaptcha']['activated'],
            'grecaptcha_site_key': config['grecaptcha']['site_key']}



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.http import HttpResponse

from swh.web.common import service
from swh.web.common.utils import reverse
from swh.web.common.utils import parse_timestamp
from swh.web.api import utils
from swh.web.api.apidoc import api_doc
from swh.web.api.apiurls import api_route
from swh.web.api.views.utils import api_lookup


def _revision_directory_by(revision, path, request_path,
                           limit=100, with_data=False):
    """
    Compute the revision matching criterion's directory or content data.

    Args:
        revision: dictionary of criterions representing a revision to lookup
        path: directory's path to lookup
        request_path: request path which holds the original context to
        limit: optional query parameter to limit the revisions log
        (default to 100). For now, note that this limit could impede the
        transitivity conclusion about sha1_git not being an ancestor of
        with_data: indicate to retrieve the content's raw data if path resolves
        to a content.

    """
    def enrich_directory_local(dir, context_url=request_path):
        return utils.enrich_directory(dir, context_url)

    rev_id, result = service.lookup_directory_through_revision(
        revision, path, limit=limit, with_data=with_data)

    content = result['content']
    if result['type'] == 'dir':  # dir_entries
        result['content'] = list(map(enrich_directory_local, content))
    elif result['type'] == 'file':  # content
        result['content'] = utils.enrich_content(content)
    elif result['type'] == 'rev':  # revision
        result['content'] = utils.enrich_revision(content)

    return result


@api_route(r'/revision/origin/(?P<origin_id>[0-9]+)'
           r'/branch/(?P<branch_name>.+)/log/',
           'api-revision-origin-log')
@api_route(r'/revision/origin/(?P<origin_id>[0-9]+)/log/',
           'api-revision-origin-log')
@api_route(r'/revision/origin/(?P<origin_id>[0-9]+)'
           r'/ts/(?P<ts>.+)/log/',
           'api-revision-origin-log')
@api_route(r'/revision/origin/(?P<origin_id>[0-9]+)'
           r'/branch/(?P<branch_name>.+)'
           r'/ts/(?P<ts>.+)/log/',
           'api-revision-origin-log')
@api_doc('/revision/origin/log/')
def api_revision_log_by(request, origin_id,
                        branch_name='HEAD',
                        ts=None):
    """
    .. http:get:: /api/1/revision/origin/(origin_id)[/branch/(branch_name)][/ts/(timestamp)]/log

        Show the commit log for a revision, searching for it based on software origin,
        branch name, and/or visit timestamp.

        This endpoint behaves like :http:get:`/api/1/revision/(sha1_git)[/prev/(prev_sha1s)]/log/`,
        but operates on the revision that has been found at a given software origin,
        close to a given point in time, pointed by a given branch.

        :param int origin_id: a software origin identifier
        :param string branch_name: optional parameter specifying a fully-qualified branch name
            associated to the software origin, e.g., "refs/heads/master". Defaults to the HEAD branch.
        :param string timestamp: optional parameter specifying a timestamp close to which the revision
            pointed by the given branch should be looked up. The timestamp can be expressed either
            as an ISO date or as a Unix one (in UTC). Defaults to now.

        :reqheader Accept: the requested response content type,
            either ``application/json`` (default) or ``application/yaml``
        :resheader Content-Type: this depends on :http:header:`Accept` header of request

        :>jsonarr object author: information about the author of the revision
        :>jsonarr string author_url: link to :http:get:`/api/1/person/(person_id)/` to get
            information about the author of the revision
        :>jsonarr object committer: information about the committer of the revision
        :>jsonarr string committer_url: link to :http:get:`/api/1/person/(person_id)/` to get
            information about the committer of the revision
        :>jsonarr string committer_date: ISO representation of the commit date (in UTC)
        :>jsonarr string date: ISO representation of the revision date (in UTC)
        :>jsonarr string directory: the unique identifier that revision points to
        :>jsonarr string directory_url: link to :http:get:`/api/1/directory/(sha1_git)/[(path)/]`
            to get information about the directory associated to the revision
        :>jsonarr string id: the revision unique identifier
        :>jsonarr boolean merge: whether or not the revision corresponds to a merge commit
        :>jsonarr string message: the message associated to the revision
        :>jsonarr array parents: the parents of the revision, i.e. the previous revisions
            that head directly to it, each entry of that array contains an unique parent
            revision identifier but also a link to :http:get:`/api/1/revision/(sha1_git)/`
            to get more information about it
        :>jsonarr string type: the type of the revision

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`, :http:method:`options`

        :statuscode 200: no error
        :statuscode 404: no revision matching the given criteria could be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`revision/origin/723566/ts/2016-01-17T00:00:00+00:00/log/`
    """ # noqa
    result = {}
    per_page = int(request.query_params.get('per_page', '10'))

    def lookup_revision_log_by_with_limit(o_id, br, ts, limit=per_page+1):
        return service.lookup_revision_log_by(o_id, br, ts, limit)

    error_msg = 'No revision matching origin %s ' % origin_id
    error_msg += ', branch name %s' % branch_name
    error_msg += (' and time stamp %s.' % ts) if ts else '.'

    rev_get = api_lookup(
        lookup_revision_log_by_with_limit, int(origin_id), branch_name, ts,
        notfound_msg=error_msg,
        enrich_fn=utils.enrich_revision)

    nb_rev = len(rev_get)
    if nb_rev == per_page+1:
        revisions = rev_get[:-1]
        last_sha1_git = rev_get[-1]['id']

        params = {k: v for k, v in {'origin_id': origin_id,
                                    'branch_name': branch_name,
                                    'ts': ts,
                                    }.items() if v is not None}

        query_params = {}
        query_params['sha1_git'] = last_sha1_git

        if request.query_params.get('per_page'):
            query_params['per_page'] = per_page

        result['headers'] = {
            'link-next': reverse('api-revision-origin-log', url_args=params,
                                 query_params=query_params)
        }

    else:
        revisions = rev_get

    result.update({'results': revisions})

    return result


@api_route(r'/revision/origin/(?P<origin_id>[0-9]+)/directory/',
           'api-revision-origin-directory')
@api_route(r'/revision/origin/(?P<origin_id>[0-9]+)/directory/(?P<path>.+)/',
           'api-revision-origin-directory')
@api_route(r'/revision/origin/(?P<origin_id>[0-9]+)'
           r'/branch/(?P<branch_name>.+)/directory/',
           'api-revision-origin-directory')
@api_route(r'/revision/origin/(?P<origin_id>[0-9]+)'
           r'/branch/(?P<branch_name>.+)/ts/(?P<ts>.+)/directory/',
           'api-revision-origin-directory')
@api_route(r'/revision/origin/(?P<origin_id>[0-9]+)'
           r'/branch/(?P<branch_name>.+)/directory/(?P<path>.+)/',
           'api-revision-origin-directory')
@api_route(r'/revision/origin/(?P<origin_id>[0-9]+)'
           r'/branch/(?P<branch_name>.+)/ts/(?P<ts>.+)'
           r'/directory/(?P<path>.+)/',
           'api-revision-origin-directory')
@api_doc('/revision/origin/directory/', tags=['hidden'])
def api_directory_through_revision_origin(request, origin_id,
                                          branch_name='HEAD',
                                          ts=None,
                                          path=None,
                                          with_data=False):
    """
    Display directory or content information through a revision identified
    by origin/branch/timestamp.
    """
    if ts:
        ts = parse_timestamp(ts)

    return _revision_directory_by({'origin_id': int(origin_id),
                                   'branch_name': branch_name,
                                   'ts': ts
                                   },
                                  path, request.path,
                                  with_data=with_data)


@api_route(r'/revision/origin/(?P<origin_id>[0-9]+)/',
           'api-revision-origin')
@api_route(r'/revision/origin/(?P<origin_id>[0-9]+)'
           r'/branch/(?P<branch_name>.+)/',
           'api-revision-origin')
@api_route(r'/revision/origin/(?P<origin_id>[0-9]+)'
           r'/branch/(?P<branch_name>.+)/ts/(?P<ts>.+)/',
           'api-revision-origin')
@api_route(r'/revision/origin/(?P<origin_id>[0-9]+)/ts/(?P<ts>.+)/',
           'api-revision-origin')
@api_doc('/revision/origin/')
def api_revision_with_origin(request, origin_id,
                             branch_name='HEAD',
                             ts=None):
    """
    .. http:get:: /api/1/revision/origin/(origin_id)/[branch/(branch_name)/][ts/(timestamp)/]

        Get information about a revision, searching for it based on software origin,
        branch name, and/or visit timestamp.

        This endpoint behaves like :http:get:`/api/1/revision/(sha1_git)/`,
        but operates on the revision that has been found at a given software origin,
        close to a given point in time, pointed by a given branch.

        :param int origin_id: a software origin identifier
        :param string branch_name: optional parameter specifying a fully-qualified branch name
            associated to the software origin, e.g., "refs/heads/master". Defaults to the HEAD branch.
        :param string timestamp: optional parameter specifying a timestamp close to which the revision
            pointed by the given branch should be looked up. The timestamp can be expressed either
            as an ISO date or as a Unix one (in UTC). Defaults to now.

        :reqheader Accept: the requested response content type,
            either ``application/json`` (default) or ``application/yaml``
        :resheader Content-Type: this depends on :http:header:`Accept` header of request

        :>json object author: information about the author of the revision
        :>json string author_url: link to :http:get:`/api/1/person/(person_id)/` to get
            information about the author of the revision
        :>json object committer: information about the committer of the revision
        :>json string committer_url: link to :http:get:`/api/1/person/(person_id)/` to get
            information about the committer of the revision
        :>json string committer_date: ISO representation of the commit date (in UTC)
        :>json string date: ISO representation of the revision date (in UTC)
        :>json string directory: the unique identifier that revision points to
        :>json string directory_url: link to :http:get:`/api/1/directory/(sha1_git)/[(path)/]`
            to get information about the directory associated to the revision
        :>json string id: the revision unique identifier
        :>json boolean merge: whether or not the revision corresponds to a merge commit
        :>json string message: the message associated to the revision
        :>json array parents: the parents of the revision, i.e. the previous revisions
            that head directly to it, each entry of that array contains an unique parent
            revision identifier but also a link to :http:get:`/api/1/revision/(sha1_git)/`
            to get more information about it
        :>json string type: the type of the revision

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`, :http:method:`options`

        :statuscode 200: no error
        :statuscode 404: no revision matching the given criteria could be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`revision/origin/13706355/branch/refs/heads/2.7/`
    """ # noqa
    return api_lookup(
        service.lookup_revision_by, int(origin_id), branch_name, ts,
        notfound_msg=('Revision with (origin_id: {}, branch_name: {}'
                      ', ts: {}) not found.'.format(origin_id,
                                                    branch_name, ts)),
        enrich_fn=utils.enrich_revision)


@api_route(r'/revision/(?P<sha1_git>[0-9a-f]+)/', 'api-revision')
@api_doc('/revision/')
def api_revision(request, sha1_git):
    """
    .. http:get:: /api/1/revision/(sha1_git)/

        Get information about a revision in the archive.
        Revisions are identified by **sha1** checksums, compatible with Git commit identifiers.
        See :func:`swh.model.identifiers.revision_identifier` in our data model module for details
        about how they are computed.

        :param string sha1_git: hexadecimal representation of the revision **sha1_git** identifier

        :reqheader Accept: the requested response content type,
            either ``application/json`` (default) or ``application/yaml``
        :resheader Content-Type: this depends on :http:header:`Accept` header of request

        :>json object author: information about the author of the revision
        :>json string author_url: link to :http:get:`/api/1/person/(person_id)/` to get
            information about the author of the revision
        :>json object committer: information about the committer of the revision
        :>json string committer_url: link to :http:get:`/api/1/person/(person_id)/` to get
            information about the committer of the revision
        :>json string committer_date: ISO representation of the commit date (in UTC)
        :>json string date: ISO representation of the revision date (in UTC)
        :>json string directory: the unique identifier that revision points to
        :>json string directory_url: link to :http:get:`/api/1/directory/(sha1_git)/[(path)/]`
            to get information about the directory associated to the revision
        :>json string id: the revision unique identifier
        :>json boolean merge: whether or not the revision corresponds to a merge commit
        :>json string message: the message associated to the revision
        :>json array parents: the parents of the revision, i.e. the previous revisions
            that head directly to it, each entry of that array contains an unique parent
            revision identifier but also a link to :http:get:`/api/1/revision/(sha1_git)/`
            to get more information about it
        :>json string type: the type of the revision

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`, :http:method:`options`

        :statuscode 200: no error
        :statuscode 400: an invalid **sha1_git** value has been provided
        :statuscode 404: requested revision can not be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`revision/aafb16d69fd30ff58afdd69036a26047f3aebdc6/`
    """ # noqa
    return api_lookup(
        service.lookup_revision, sha1_git,
        notfound_msg='Revision with sha1_git {} not found.'.format(sha1_git),
        enrich_fn=utils.enrich_revision)


@api_route(r'/revision/(?P<sha1_git>[0-9a-f]+)/raw/',
           'api-revision-raw-message')
@api_doc('/revision/raw/', tags=['hidden'], handle_response=True)
def api_revision_raw_message(request, sha1_git):
    """Return the raw data of the message of revision identified by sha1_git
    """
    raw = service.lookup_revision_message(sha1_git)
    response = HttpResponse(raw['message'],
                            content_type='application/octet-stream')
    response['Content-disposition'] = \
        'attachment;filename=rev_%s_raw' % sha1_git
    return response


@api_route(r'/revision/(?P<sha1_git>[0-9a-f]+)/directory/',
           'api-revision-directory')
@api_route(r'/revision/(?P<sha1_git>[0-9a-f]+)/directory/(?P<dir_path>.+)/',
           'api-revision-directory')
@api_doc('/revision/directory/')
def api_revision_directory(request, sha1_git,
                           dir_path=None,
                           with_data=False):
    """
    .. http:get:: /api/1/revision/(sha1_git)/directory/[(path)/]

        Get information about directory (entry) objects associated to revisions.
        Each revision is associated to a single "root" directory.
        This endpoint behaves like :http:get:`/api/1/directory/(sha1_git)/[(path)/]`,
        but operates on the root directory associated to a given revision.

        :param string sha1_git: hexadecimal representation of the revision **sha1_git** identifier
        :param string path: optional parameter to get information about the directory entry
            pointed by that relative path

        :reqheader Accept: the requested response content type,
            either ``application/json`` (default) or ``application/yaml``
        :resheader Content-Type: this depends on :http:header:`Accept` header of request

        :>json array content: directory entries as returned by :http:get:`/api/1/directory/(sha1_git)/[(path)/]`
        :>json string path: path of directory from the revision root one
        :>json string revision: the unique revision identifier
        :>json string type: the type of the directory

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`, :http:method:`options`

        :statuscode 200: no error
        :statuscode 400: an invalid **sha1_git** value has been provided
        :statuscode 404: requested revision can not be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`revision/f1b94134a4b879bc55c3dacdb496690c8ebdc03f/directory/`
    """ # noqa
    return _revision_directory_by({'sha1_git': sha1_git},
                                  dir_path, request.path,
                                  with_data=with_data)


@api_route(r'/revision/(?P<sha1_git>[0-9a-f]+)/log/', 'api-revision-log')
@api_route(r'/revision/(?P<sha1_git>[0-9a-f]+)'
           r'/prev/(?P<prev_sha1s>[0-9a-f/]+)/log/',
           'api-revision-log')
@api_doc('/revision/log/')
def api_revision_log(request, sha1_git, prev_sha1s=None):
    """
    .. http:get:: /api/1/revision/(sha1_git)[/prev/(prev_sha1s)]/log/

        Get a list of all revisions heading to a given one, in other words show the commit log.

        :param string sha1_git: hexadecimal representation of the revision **sha1_git** identifier
        :param string prev_sha1s: optional parameter representing the navigation breadcrumbs
            (descendant revisions previously visited). If multiple values, use / as delimiter.
            If provided, revisions information will be added at the beginning of the returned list.
        :query int per_page: number of elements in the returned list, for pagination purpose

        :reqheader Accept: the requested response content type,
            either ``application/json`` (default) or ``application/yaml``
        :resheader Content-Type: this depends on :http:header:`Accept` header of request
        :resheader Link: indicates that a subsequent result page is available and contains
            the url pointing to it

        :>jsonarr object author: information about the author of the revision
        :>jsonarr string author_url: link to :http:get:`/api/1/person/(person_id)/` to get
            information about the author of the revision
        :>jsonarr object committer: information about the committer of the revision
        :>jsonarr string committer_url: link to :http:get:`/api/1/person/(person_id)/` to get
            information about the committer of the revision
        :>jsonarr string committer_date: ISO representation of the commit date (in UTC)
        :>jsonarr string date: ISO representation of the revision date (in UTC)
        :>jsonarr string directory: the unique identifier that revision points to
        :>jsonarr string directory_url: link to :http:get:`/api/1/directory/(sha1_git)/[(path)/]`
            to get information about the directory associated to the revision
        :>jsonarr string id: the revision unique identifier
        :>jsonarr boolean merge: whether or not the revision corresponds to a merge commit
        :>jsonarr string message: the message associated to the revision
        :>jsonarr array parents: the parents of the revision, i.e. the previous revisions
            that head directly to it, each entry of that array contains an unique parent
            revision identifier but also a link to :http:get:`/api/1/revision/(sha1_git)/`
            to get more information about it
        :>jsonarr string type: the type of the revision

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`, :http:method:`options`

        :statuscode 200: no error
        :statuscode 400: an invalid **sha1_git** value has been provided
        :statuscode 404: requested revision can not be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`revision/e1a315fa3fa734e2a6154ed7b5b9ae0eb8987aad/log/`
    """ # noqa
    result = {}
    per_page = int(request.query_params.get('per_page', '10'))

    def lookup_revision_log_with_limit(s, limit=per_page+1):
        return service.lookup_revision_log(s, limit)

    error_msg = 'Revision with sha1_git %s not found.' % sha1_git
    rev_get = api_lookup(lookup_revision_log_with_limit, sha1_git,
                         notfound_msg=error_msg,
                         enrich_fn=utils.enrich_revision)

    nb_rev = len(rev_get)
    if nb_rev == per_page+1:
        rev_backward = rev_get[:-1]
        new_last_sha1 = rev_get[-1]['id']
        query_params = {}

        if request.query_params.get('per_page'):
            query_params['per_page'] = per_page

        result['headers'] = {
            'link-next': reverse('api-revision-log',
                                 url_args={'sha1_git': new_last_sha1},
                                 query_params=query_params)
        }

    else:
        rev_backward = rev_get

    if not prev_sha1s:  # no nav breadcrumbs, so we're done
        revisions = rev_backward

    else:
        rev_forward_ids = prev_sha1s.split('/')
        rev_forward = api_lookup(
            service.lookup_revision_multiple, rev_forward_ids,
            notfound_msg=error_msg,
            enrich_fn=utils.enrich_revision)
        revisions = rev_forward + rev_backward

    result.update({
        'results': revisions
    })
    return result



#!/usr/bin/env python3
# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License v3 or later
# See top-level LICENSE file for more information

from setuptools import setup, find_packages

from os import path
from io import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


def parse_requirements(name=None):
    if name:
        reqf = 'requirements-%s.txt' % name
    else:
        reqf = 'requirements.txt'

    requirements = []
    if not path.exists(reqf):
        return requirements

    with open(reqf) as f:
        for line in f.readlines():
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            requirements.append(line)
    return requirements


setup(
    name='swh.web',
    description='Software Heritage Web UI',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Software Heritage developers',
    author_email='swh-devel@inria.fr',
    url='https://forge.softwareheritage.org/diffusion/DWUI/',
    packages=find_packages(),
    scripts=[],
    install_requires=parse_requirements() + parse_requirements('swh'),
    setup_requires=['vcversioner'],
    extras_require={'testing': parse_requirements('test')},
    vcversioner={},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",  # noqa
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
    ],
    project_urls={
        'Bug Reports': 'https://forge.softwareheritage.org/maniphest',
        'Funding': 'https://www.softwareheritage.org/donate',
        'Source': 'https://forge.softwareheritage.org/source/swh-web',
    },
)
